
/* Copyright (c) Mark J. Kilgard, 1994.  */

/* This program is freely distributable without licensing fees 
   and is provided without guarantee or warrantee expressed or 
   implied. This program is -not- in the public domain. */

/* New GLUT 3.0 glutGetModifiers() functionality used to make Shift-Left
   mouse scale the dinosaur's size. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>       /* for cos(), sin(), and sqrt() */
#include <GL/glu.h>
#include <GL/glut.h>
#include "trackball.h"

#ifndef M_PI
#define M_PI   3.14159265358979323846
#endif

typedef enum {
  RESERVED, DINOSAUR
} displayLists;

int spinning = 0, moving = 0;
int beginx, beginy;
int W = 300, H = 300;
float curquat[4];
float lastquat[4];
int newModel = 1;
int scaling;
float scalefactor = 1.0;

GLfloat lightZeroPosition[] = {10.0, 4.0, 10.0, 1.0};
GLfloat lightZeroColor[] = {0.8, 1.0, 0.8, 1.0}; /* green-tinted */
GLfloat lightOnePosition[] = {-1.0, -2.0, 1.0, 0.0};
GLfloat lightOneColor[] = {0.6, 0.3, 0.2, 1.0}; /* red-tinted */

GLfloat skinColor[] = {0.1, 1.0, 0.1, 1.0};
GLfloat underColor[] = {0.05, 0.5, 0.05, 1.0};

static GLuint texName;

#define MAX_SIZE 40
#define MAX_RANDOM 0.01
#define MAX_RANGE 1.0
#define MISSING 0.0
double z[MAX_SIZE][MAX_SIZE];
double Nx[MAX_SIZE][MAX_SIZE];
double Ny[MAX_SIZE][MAX_SIZE];
double Nz[MAX_SIZE][MAX_SIZE];

double x ( int i)
{
	return 2.0 * i / (MAX_SIZE - 1) - 1.0;
}

double y ( int i)
{
	return x ( i );
}

void Normalize ( int i, int j )
{
	double norm;
	double step = x(1);

	if ( (i == MAX_SIZE - 1) || (j == MAX_SIZE - 1) )
	{
		Nx[i][j] = 0.0;
		Ny[i][j] = 0.0;
		Nz[i][j] = 1.0;
		return;
	}

	Nx[i][j] = - step * ( z[i+1][j] - z[i][j] );
	Ny[i][j] = - step * ( z[i][j+1] - z[i][j] );
	Nz[i][j] = step * step;

	norm = Nx[i][j]*Nx[i][j] + Ny[i][j]*Ny[i][j] + Nz[i][j]*Nz[i][j];
	norm = sqrt ( norm );

	Nx[i][j] /= norm;
	Ny[i][j] /= norm;
	Nz[i][j] /= norm;
}

void LoadData ( double Range, double Theta, double Radius,
        double WedgeDepth, double WedgeTheta, double WedgeAcute )
{
	int i, j;
	double a, b, scale, Wa, Wb, Ws, X, Y, NoWedge, WithWedge;

	if ( Radius < 0.0 )
		Radius = 0.0;
	else if ( Radius > 1.0 )
		Radius = 1.0;

	if ( Range < 0.0 )
		Range = 0.0;
	else if ( Range > MAX_RANGE )
		Range = MAX_RANGE;

	if ( WedgeAcute <= 0.001 )
	{
		WedgeAcute = 0.001;
		WedgeDepth = 0.0;
	}
	if ( WedgeAcute >= M_PI / 3.0 )
		WedgeAcute = M_PI / 3.0;

	if ( WedgeDepth < 0.0 )
		WedgeDepth = 0.0;
	else if ( WedgeDepth > Range )
		WedgeDepth = Range;

	b = sin ( Theta );
	a = cos ( Theta );
	scale = Range / (2*Radius);
	Wa = sin ( WedgeTheta );
	Wb = cos ( WedgeTheta );
	Ws = 1.0 / tan ( WedgeAcute );

	for ( i = 0; i < MAX_SIZE; i++ )
		for ( j = 0; j < MAX_SIZE; j++ )
		{
			X = x(i);
			Y = y(j);
			if ( X*X + Y*Y > Radius * Radius )
				z[i][j] = MISSING;
			else
			{
				NoWedge = (a*X + b*Y) * scale + scale;
				WithWedge = Ws * fabs ( Wa*X + Wb*Y ) +
					Range - WedgeDepth;
					
				z[i][j] = NoWedge < WithWedge? NoWedge: WithWedge;
			}
		}
	for ( i = 0; i < MAX_SIZE; i++ )
		for ( j = 0; j < MAX_SIZE; j++ )
		{
			double temp = drand48();
			#ifdef __QNX__
			temp = 0.5;
			#endif
			temp = 2*MAX_RANDOM * temp - MAX_RANDOM;
			z[i][j] += temp;
		}
	for ( i = 0; i < MAX_SIZE; i++ )
		for ( j = 0; j < MAX_SIZE; j++ )
			Normalize ( i, j );
}

void
makeDinosaur(void)
{
	int i, j;
	double X0, X1, Y;

	LoadData ( 1.0, M_PI/2.0, 1.0, 0.67, M_PI/4.0, M_PI/12.0 );
	
	glNewList(DINOSAUR, GL_COMPILE);
	
	//glMaterialfv ( GL_FRONT, GL_DIFFUSE, skinColor );
	//glMaterialfv ( GL_BACK, GL_DIFFUSE, underColor );

	glBindTexture ( GL_TEXTURE_1D, texName );

    for ( i = 0; i < MAX_SIZE - 1; i++ )
	{
		X0 = x(i);
		X1 = x(i+1);
		glBegin ( GL_QUAD_STRIP );
        for ( j = 0; j < MAX_SIZE; j++ )
		{
			Y = y(j);
			glNormal3d ( Nx[i][j], Ny[i][j], Nz[i][j] );
            glVertex3d ( X0, Y, z[i][j] );
            glVertex3d ( X1, Y, z[i+1][j] );
		}
		glEnd();
	}
	glEndList();
}

#define stripeImageWidth 32
GLubyte stripeImage[4*stripeImageWidth];

//static GLfloat xequalzero[] = { 1.0, 0.0, 0.0, 0.0};
static GLfloat xequalzero[] = { 0.0, 0.0, 1.0, 0.0};
static GLfloat* currentCoeff;
static GLenum currentPlane;
static GLint currentGenMode;

void makeStripeImage1()
{
	int i;

	/* 1, 2, 3, 4 have 255, 0, 0, 255 */
	/* rest have 0, 255, 0, 255 */
	for ( i = 0; i < stripeImageWidth; i++ )
	{
		stripeImage[4*i] = (GLubyte) (  0 );
		stripeImage[4*i+1] = (GLubyte) (  255 );
		stripeImage[4*i+2] = (GLubyte) 0;
		stripeImage[4*i+3] = (GLubyte) 255;
	}
	for ( i = 3; i < stripeImageWidth; i += 4 )
	{
		stripeImage[4*i] = (GLubyte) ( 255  );
		stripeImage[4*i+1] = (GLubyte) (  0 );
		stripeImage[4*i+2] = (GLubyte) 0;
		stripeImage[4*i+3] = (GLubyte) 255;
	}
}

void makeStripeImage2()
{
	int i;
	i = 0;
		stripeImage[4*i] = (GLubyte) 0;
		stripeImage[4*i+1] = (GLubyte) 0;
		stripeImage[4*i+2] = (GLubyte) 0;
		stripeImage[4*i+3] = (GLubyte) 255;
	for ( i = 1; i < stripeImageWidth - 1; i++ )
	{
		stripeImage[4*i] = (GLubyte) ( 256 - ( 8 * i ) );
		stripeImage[4*i+1] = (GLubyte) ( 8 * i  );
		stripeImage[4*i+2] = (GLubyte) 0;
		stripeImage[4*i+3] = (GLubyte) 255;
	}
	i = 31;
		stripeImage[4*i] = (GLubyte) 0;
		stripeImage[4*i+1] = (GLubyte) 225;
		stripeImage[4*i+2] = (GLubyte) 0;
		stripeImage[4*i+3] = (GLubyte) 255;
}

double oneThird = 1.0/3.0;
double twoThirds = 2.0/3.0;

double red ( double x )
{
	if ( x <= 0.2 )
		return (0.2 - x )*5.0;
	if ( x <= 0.6 )
		return 0.0;
	if ( x <= 0.8 )
		return (x - 0.6)*5.0;
	return 1.0;
}

double blue ( double x )
{
	if ( x <= 0.4 )
		return  1.0;
	if ( x <= 0.6 )
		return (0.6 - x)*5.0;
	return 0.0;
}

double green ( double x )
{
	if ( x <= 0.2 )
		return 0.0;
	if ( x <= 0.4 )
		return (x - 0.2)*5.0;
	if ( x <= 0.8 )
		return 1.0;
	return (1.0 - x )*5.0;
}

//void makeStripeImage3()
void makeStripeImage()
{
	int i;
	double range = stripeImageWidth - 1;

	for ( i = 1; i < stripeImageWidth - 1; i++ )
	{
		stripeImage[4*i] = (GLubyte) (255 * red ( i / range ));
		stripeImage[4*i+1] = (GLubyte) (255 * green ( i /range ));
		stripeImage[4*i+2] = (GLubyte) (255 *blue ( i/ range ));
		stripeImage[4*i+3] = (GLubyte) 255;
	}
	
}

void initTexture()
{
	makeStripeImage();
	glPixelStorei ( GL_UNPACK_ALIGNMENT, 1 );
	glGenTextures ( 1, &texName );
	glBindTexture ( GL_TEXTURE_1D, texName );

	glTexParameteri ( GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri ( GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri ( GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexImage1D ( GL_TEXTURE_1D, 0, GL_RGBA, stripeImageWidth, 0,
		GL_RGBA, GL_UNSIGNED_BYTE, stripeImage );
	glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	currentCoeff = xequalzero;
	currentGenMode = GL_OBJECT_LINEAR;
	currentPlane = GL_OBJECT_PLANE;
	glTexGeni ( GL_S, GL_TEXTURE_GEN_MODE, currentGenMode );
	glTexGenfv ( GL_S, currentPlane, currentCoeff );
	glEnable ( GL_TEXTURE_GEN_S );
	glEnable ( GL_TEXTURE_1D );
}

void
recalcModelView(void)
{
  GLfloat m[4][4];

  glPopMatrix();
  glPushMatrix();
  build_rotmatrix(m, curquat);
  glMultMatrixf(&m[0][0]);
  glScalef(scalefactor, scalefactor, scalefactor);
  newModel = 0;
}

void
showMessage(GLfloat x, GLfloat y, GLfloat z, char *message)
{
  glPushMatrix();
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_1D);
  glTranslatef(x, y, z);
  glScalef(.002, .002, .002);
  while (*message) {
    glutStrokeCharacter(GLUT_STROKE_ROMAN, *message);
    message++;
  }
  glEnable(GL_LIGHTING);
  glEnable(GL_TEXTURE_1D);
  glPopMatrix();
}

void
redraw(void)
{
  if (newModel)
    recalcModelView();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glCallList(DINOSAUR);
  //showMessage(2, 7.1, 4.1, "Spin me.");
  showMessage(-0.5, 0, 1, "Spin me.");
  glutSwapBuffers();
}

void
myReshape(int w, int h)
{
  glViewport(0, 0, w, h);
  W = w;
  H = h;
}

void
mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    spinning = 0;
    glutIdleFunc(NULL);
    moving = 1;
    beginx = x;
    beginy = y;
    if (glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
      scaling = 1;
    } else {
      scaling = 0;
    }
  }
  if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
    moving = 0;
  }
}

void
animate(void)
{
  add_quats(lastquat, curquat, curquat);
  newModel = 1;
  glutPostRedisplay();
}

void
motion(int x, int y)
{
  if (scaling) {
    scalefactor = scalefactor * (1.0 + (((float) (beginy - y)) / H));
    beginx = x;
    beginy = y;
    newModel = 1;
    glutPostRedisplay();
    return;
  }
  if (moving) {
    trackball(lastquat,
      (2.0 * beginx - W) / W,
      (H - 2.0 * beginy) / H,
      (2.0 * x - W) / W,
      (H - 2.0 * y) / H
      );
    beginx = x;
    beginy = y;
    spinning = 1;
    glutIdleFunc(animate);
  }
}

GLboolean lightZeroSwitch = GL_TRUE, lightOneSwitch = GL_TRUE;

void
controlLights(int value)
{
  switch (value) {
  case 1:
    lightZeroSwitch = !lightZeroSwitch;
    if (lightZeroSwitch) {
      glEnable(GL_LIGHT0);
    } else {
      glDisable(GL_LIGHT0);
    }
    break;
  case 2:
    lightOneSwitch = !lightOneSwitch;
    if (lightOneSwitch) {
      glEnable(GL_LIGHT1);
    } else {
      glDisable(GL_LIGHT1);
    }
    break;
#ifdef GL_MULTISAMPLE_SGIS
  case 3:
    if (glIsEnabled(GL_MULTISAMPLE_SGIS)) {
      glDisable(GL_MULTISAMPLE_SGIS);
    } else {
      glEnable(GL_MULTISAMPLE_SGIS);
    }
    break;
#endif
  case 4:
    glutFullScreen();
    break;
  case 5:
    exit(0);
    break;
  }
  glutPostRedisplay();
}

void
vis(int visible)
{
  if (visible == GLUT_VISIBLE) {
    if (spinning)
      glutIdleFunc(animate);
  } else {
    if (spinning)
      glutIdleFunc(NULL);
  }
}

int
main(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
  trackball(curquat, 0.0, 0.0, 0.0, 0.0);
  glutCreateWindow("demospin");
  glutDisplayFunc(redraw);
  glutReshapeFunc(myReshape);
  glutVisibilityFunc(vis);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutCreateMenu(controlLights);
  glutAddMenuEntry("Toggle right light", 1);
  glutAddMenuEntry("Toggle left light", 2);
  if (glutGet(GLUT_WINDOW_NUM_SAMPLES) > 0) {
    glutAddMenuEntry("Toggle multisampling", 3);
    glutSetWindowTitle("demospin (multisample capable)");
  }
  glutAddMenuEntry("Full screen", 4);
  glutAddMenuEntry("Quit", 5);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
	initTexture();
  makeDinosaur();
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glMatrixMode(GL_PROJECTION);
  gluPerspective( /* field of view in degree */ 40.0,
  /* aspect ratio */ 1.0,
    /* Z near */ 1.0, /* Z far */ 6.0);
  glMatrixMode(GL_MODELVIEW);
  gluLookAt(0.0, 0.0, 4.0,  /* eye is at (0,0,30) */
    0.0, 0.0, 0.0,      /* center is at (0,0,0) */
    0.0, 1.0, 0.);      /* up is in positive Y direction */
  glPushMatrix();       /* dummy push so we can pop on model
                           recalc */
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightZeroColor);
  glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05);
  glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glLineWidth(2.0);
  glutMainLoop();
  return 0;             /* ANSI C requires main to return int. */
}
